package com.lanren.huhu.partner.schedule;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.lanren.huhu.partner.domain.*;
import com.lanren.huhu.partner.model.ParentAgent;
import com.lanren.huhu.partner.model.ParentPartner;
import com.lanren.huhu.partner.service.*;
import com.lanren.huhu.partner.util.DateUtils;
import lombok.Data;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.*;

import static com.lanren.huhu.partner.constants.Constants.AGENT_LEVEL_1;

/**
 * @author chen
 * @title: AgentDailyExpandTask
 * @projectName partner
 * @description: 代理商每日拓展用户数,  按用户的代理商层次查找归属
 * @package com.lanren.huhu.partner.schedule
 * @date 2019-07-08 15:50
 */
@Component
public class AgentDailyExpandTask {
    private static Logger logger = LoggerFactory.getLogger(AgentDailyExpandTask.class);
    private static final int PAGE_SIZE = 5000;
    /**
     * 用来遍历的表
     * key: uid
     * valye: partner_level
     */
    private Map<Integer, Integer> partnerLevelMap = new HashMap<Integer, Integer>();
    /**
     * key: uid
     * value: invite_uid
     */
    private Map<Integer, Integer> partnerInviteMap = new HashMap<Integer, Integer>();
    /**
     * key: uid
     * value: agent_id
     */
    private Map<Integer, Integer> partnerAgentMap = new HashMap<Integer, Integer>();
    /**
     * key: uid
     * value: reg_time
     */
    private Map<Integer, Date> partnerRegMap = new HashMap<Integer, Date>();

    /**
     * key: agent_id
     * value: uid
     */
    private Map<Integer, Integer> agentPartnerMap = new HashMap<Integer, Integer>();
    /**
     * key: agent_id
     * value: 层级上级agent_id
     */
    private Map<Integer, Integer> agentParentMap = new HashMap<Integer, Integer>();
    /**
     * key: agent_id
     * value: agent_level
     */
    private Map<Integer, Integer> agentLevelMap = new HashMap<Integer, Integer>();

    @Autowired
    PartnerInviteRelationService partnerInviteRelationService;
    @Autowired
    PartnerAccountService partnerAccountService;
    @Autowired
    UserAgentService userAgentService;
    @Autowired
    UserInfoService userInfoService;
    @Autowired
    AgentExpandDetailService agentExpandDetailService;
    @Autowired
    AgentExpandDailyService agentExpandDailyService;

    private void initPartnerRegMap() {
        try {
            partnerRegMap.clear();
            int pageNo = 1;
            int total = 0;
            int totalPage = 0;
            int cnt = 0;
            while (true) {
                Page<UserInfo> page = new Page<UserInfo>(pageNo, PAGE_SIZE);
                IPage<UserInfo> ipage = userInfoService.selectPageByTs(page);
                if (total == 0) {
                    total = (int) ipage.getTotal();
                    totalPage = total % PAGE_SIZE == 0 ? total / PAGE_SIZE : total / PAGE_SIZE + 1;
                    logger.info("totalPage is {} ", totalPage);
                }
                List<UserInfo> userInfoList = ipage.getRecords();
                if (userInfoList != null && userInfoList.size() > 0) {
                    for (UserInfo userInfo : userInfoList) {
                        partnerRegMap.put(userInfo.getUserId(), userInfo.getRegisterTime());
                        cnt++;
                    }
                }
                if (pageNo >= totalPage) {
                    break;
                }
                pageNo++;
            }
            logger.info("partnerRegMap cnt is {}", cnt);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }
    }

    private void initPartnerLevelMap() {
        try {
            partnerLevelMap.clear();
            int pageNo = 1;
            int total = 0;
            int totalPage = 0;
            int cnt = 0;
            while (true) {
                Page<PartnerAccount> page = new Page<PartnerAccount>(pageNo, PAGE_SIZE);
                IPage<PartnerAccount> ipage = partnerAccountService.selectPageByTs(page);
                if (total == 0) {
                    total = (int) ipage.getTotal();
                    totalPage = total % PAGE_SIZE == 0 ? total / PAGE_SIZE : total / PAGE_SIZE + 1;
                    logger.info("totalPage is {} ", totalPage);
                }
                List<PartnerAccount> partnerAccountList = ipage.getRecords();
                if (partnerAccountList != null && partnerAccountList.size() > 0) {
                    for (PartnerAccount partnerAccount : partnerAccountList) {
                        Integer partnerLevel = partnerAccount.getIsSuperPartner() == 1 ? 30 : partnerAccount.getPartnerLevel();
                        partnerLevelMap.put(partnerAccount.getUserId(), partnerLevel);
                        cnt++;
                    }
                }
                if (pageNo >= totalPage) {
                    break;
                }
                pageNo++;
            }
            logger.info("partnerLevelMap cnt is {}", cnt);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }
    }

    private void initPartnerInviteMap() {
        try {
            partnerInviteMap.clear();
            int pageNo = 1;
            int total = 0;
            int totalPage = 0;
            int cnt = 0;
            while (true) {
                Page<PartnerInviteRelation> page = new Page<PartnerInviteRelation>(pageNo, PAGE_SIZE);
                IPage<PartnerInviteRelation> ipage = partnerInviteRelationService.selectPageByTs(page);
                if (total == 0) {
                    total = (int) ipage.getTotal();
                    totalPage = total % PAGE_SIZE == 0 ? total / PAGE_SIZE : total / PAGE_SIZE + 1;
                    logger.info("totalPage is {} ", totalPage);
                }
                List<PartnerInviteRelation> partnerInviteRelationList = ipage.getRecords();
                if (partnerInviteRelationList != null && partnerInviteRelationList.size() > 0) {
                    for (PartnerInviteRelation partnerInviteRelation : partnerInviteRelationList) {
                        if (partnerInviteRelation.getInviteUserId() > 0) {
                            partnerInviteMap.put(partnerInviteRelation.getUserId(), partnerInviteRelation.getInviteUserId());
                        }
                        cnt++;
                    }
                }
                if (pageNo >= totalPage) {
                    break;
                }
                pageNo++;
            }
            logger.info("partnerInviteMap cnt is {}", cnt);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }
    }

    private void initAgentMap() {
        try {
            partnerAgentMap.clear();
            agentPartnerMap.clear();
            agentLevelMap.clear();
            agentParentMap.clear();
            int pageNo = 1;
            int total = 0;
            int totalPage = 0;
            int cnt = 0;
            while (true) {
                Page<UserAgent> page = new Page<UserAgent>(pageNo, PAGE_SIZE);
                IPage<UserAgent> ipage = userAgentService.selectPageByTs(page);
                if (total == 0) {
                    total = (int) ipage.getTotal();
                    totalPage = total % PAGE_SIZE == 0 ? total / PAGE_SIZE : total / PAGE_SIZE + 1;
                    logger.info("totalPage is {} ", totalPage);
                }
                List<UserAgent> userAgentList = ipage.getRecords();
                if (userAgentList != null && userAgentList.size() > 0) {
                    for (UserAgent userAgent : userAgentList) {
                        Integer agentLevel = userAgent.getAgentLevel();
                        partnerAgentMap.put(userAgent.getUserId(), userAgent.getAgentId());
                        agentPartnerMap.put(userAgent.getAgentId(), userAgent.getUserId());
                        agentLevelMap.put(userAgent.getAgentId(), agentLevel);
                        if (userAgent.getParentAgentId() > 0) {
                            agentParentMap.put(userAgent.getAgentId(), userAgent.getParentAgentId());
                        }
                        cnt++;
                    }
                }
                if (pageNo >= totalPage) {
                    break;
                }
                pageNo++;
            }
            logger.info("agent cnt is {}", cnt);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }
    }


    public List<ParentAgent> getAgentListByUserId(Integer userId) {
        if (logger.isInfoEnabled()) {
            logger.info("getAgentListByUserId {} for user: ......", userId);
        }
        ArrayList<ParentAgent> agentList = new ArrayList<ParentAgent>();
        HashSet<Integer> existingUserIdSet = new HashSet<Integer>();
        /**
         * 无限找
         * 先无限找上级合伙人, 直到找到第一个是代理商的合伙人
         */

        /**
         * 要改成从自己开始找 所以添加判断 看自己是不是
         */
        InnerUserAgent selfAgent = getUserAgentByUserId(userId);
        /**
         *   代理商关系包含自己的条件:
         *   1. 层级关系 都包含自己
         *   2. 推荐关系 只有当自己是总代时才包含自己
         */

        /**
         * 1.4.5更新: 平推奖励都不包含自己了
         */
        int loopUserId = userId;
        int loopTimes = 0;
        InnerUserAgent firstAgent = null;
        /**
         * 自己不是代理商的话 按C端邀请关系找到第一个代理商
         * 自己是代理商的话 按B端找, 如果是找平推, 最后都找完再剔除自己
         */
        if (selfAgent == null) {
            /**
             * 自己不是代理商时 按C端邀请关系查找代理商
             */
            while (true) {
                if (logger.isDebugEnabled()) {
                    logger.info("now loop uid is {}", loopUserId);
                }
                existingUserIdSet.add(loopUserId);
                ParentPartner parentPartner = getParentPartner(loopUserId);
                if (parentPartner != null) {
                    InnerUserAgent agent = getUserAgentByUserId(parentPartner.getUserId());
                    if (agent != null) {
                        firstAgent = agent;
                        break;
                    } else {
                        /**
                         * 这个人不是代理商 继续往上找
                         */
                        loopUserId = parentPartner.getUserId();
                        if (existingUserIdSet.contains(loopUserId)) {
                            logger.info("发现互为上级的用户邀请关系 用户ID({})", loopUserId);
                            logger.info("退出循环, 不再继续查找");
                            break;
                        }
                    }
                } else {
                    /**
                     * 没有上级合伙人 打印结束信息 跳出循环
                     */
                    if (logger.isDebugEnabled()) {
                        logger.info("loop end with userId: {}", loopUserId);
                    }
                    break;
                }
                loopTimes++;
                if (loopTimes > 500) {
                    agentList.clear();
                    logger.error("用户ID:{} 合伙人关系异常, 已超500层", userId);
                    break;
                }
            }
        } else {
            firstAgent = selfAgent;
        }

        /**
         * 找到了第一个是代理商的合伙人
         */
        if (null != firstAgent) {
            /**
             * 然后开始 无限找 这个人的上级代理商
             */
            if (logger.isDebugEnabled()) {
                logger.info("find first agent({}) for user: {}", firstAgent.getAgentId(), userId);
            }
            /**
             * 添加代理到list时 都需要加个判断 如果找到总代 就不再找了
             */
            agentList.add(new ParentAgent(firstAgent.getUserId(), firstAgent.getAgentId(), firstAgent.getAgentLevel()));
            if (firstAgent.getAgentLevel() != AGENT_LEVEL_1) {
                if (logger.isDebugEnabled()) {
                    logger.info("loop find parent agent start ......");
                }
                int loogAgentId = firstAgent.getAgentId();
                loopTimes = 0;
                while (true) {
                    ParentAgent loopUserAgent = getParentAgent(loogAgentId);
                    if (null == loopUserAgent) {
                        break;
                    } else {
                        if (logger.isDebugEnabled()) {
                            logger.info("parent agent is ", loopUserAgent.getAgentId());
                        }
                        if (agentList.contains(loopUserAgent) || loopUserAgent.getAgentId() == firstAgent.getAgentId()) {
                            logger.info("发现互为上级的代理关系 代理商({})", loopUserAgent.getAgentId());
                            logger.info("退出循环, 不再继续查找");
                            break;
                        }
                        /**
                         * 添加代理到list时 都需要加个判断 如果找到总代 就不再找了
                         */
                        agentList.add(new ParentAgent(loopUserAgent.getUserId(), loopUserAgent.getAgentId(), loopUserAgent.getLevel()));
                        if (loopUserAgent.getLevel() == AGENT_LEVEL_1) {
                            break;
                        }
                        loogAgentId = loopUserAgent.getAgentId();
                    }
                    loopTimes++;
                    if (loopTimes > 500) {
                        agentList.clear();
                        logger.error("用户ID:{} 合伙人关系异常, 已超500层", userId);
                        break;
                    }
                }
                /**
                 * 找完 "第一个是代理商的合伙人" 的 全部上级代理后 退出while循环
                 */
                if (logger.isDebugEnabled()) {
                    logger.info("loop find parent agent end");
                }
            } else {
                if (logger.isInfoEnabled()) {
                    logger.info("已找到总代 不再往下找, 总代id: {}", firstAgent.getAgentId());
                }
            }
//            /**
//             * 如果是平推奖励 过滤掉自己
//             */
//            if (parentColumnName.equals(PARENT_COLUMN_NAME_PINGTUI)
//                    && selfAgent != null
//                    && agentList.size() > 0) {
//                agentList.remove(0);
//            }
        }

        if (logger.isInfoEnabled()) {
            logger.info("setAgentList {} for user: {} end", userId);
        }
        return agentList;
    }


    private InnerUserAgent getUserAgentByUserId(Integer userId) {
        InnerUserAgent innerUserAgent = null;
        if (partnerAgentMap.containsKey(userId)) {
            innerUserAgent = new InnerUserAgent();
            innerUserAgent.setUserId(userId);
            innerUserAgent.setAgentId(partnerAgentMap.get(userId));
            innerUserAgent.setAgentLevel(agentLevelMap.get(innerUserAgent.getAgentId()));
        }
        return innerUserAgent;
    }

    private ParentPartner getParentPartner(Integer userId) {
        ParentPartner parentPartner = null;
        if (partnerInviteMap.containsKey(userId)) {
            Integer parentUid = partnerInviteMap.get(userId);
            Integer parentLevel = partnerLevelMap.get(parentUid);
            parentPartner = new ParentPartner(parentUid, parentLevel);
        }
        return parentPartner;
    }

    private ParentAgent getParentAgent(Integer agentId) {
        ParentAgent parentAgent = null;
        if (agentParentMap.containsKey(agentId)) {
            Integer parentAgentId = agentParentMap.get(agentId);
            Integer parentUserId = agentPartnerMap.get(parentAgentId);
            Integer parentAgentLevel = agentLevelMap.get(parentAgentId);
            parentAgent = new ParentAgent(parentUserId, parentAgentId, parentAgentLevel);
        }
        return parentAgent;
    }

    public void runExpandPartner() {
        /**
         * 初始化 查询用的map表
         */
        initPartnerInviteMap();
        initPartnerLevelMap();
        initPartnerRegMap();
        initAgentMap();
        /**
         * 遍历表里第一个map
         * 刷新detail数据
         */
        int now = (int) (System.currentTimeMillis() / 1000L);
        int loopTimes = 0;
        List<AgentExpandDetail> insertList = new ArrayList<AgentExpandDetail>();
        QueryWrapper<AgentExpandDetail> wrapper = new QueryWrapper<AgentExpandDetail>();
        wrapper.le("dat", DateUtils.getDate());
        agentExpandDetailService.remove(wrapper);
        for (Integer userId : partnerInviteMap.keySet()) {
            loopTimes++;
            List<ParentAgent> parentAgentList = getAgentListByUserId(userId);
            Date regTime = partnerRegMap.get(userId);
            for (ParentAgent parentAgent : parentAgentList) {
                AgentExpandDetail detail = new AgentExpandDetail();
                detail.setDat(regTime);
                detail.setAgentId(parentAgent.getAgentId());
                detail.setUserId(parentAgent.getUserId());
                detail.setExpandUserId(userId);
                detail.setExpandUserRegTime(regTime);
                detail.setCreatedAt(now);
                detail.setUpdatedAt(now);
                insertList.add(detail);
            }
            if (loopTimes >= 2000) {
                agentExpandDetailService.batchInsert(insertList);
                insertList.clear();
                loopTimes = 0;
            }
        }
        if (loopTimes > 0) {
            agentExpandDetailService.batchInsert(insertList);
            insertList.clear();
        }
        /**
         * 更新daily数据
         */
        QueryWrapper<AgentExpandDaily> wrapper2 = new QueryWrapper<AgentExpandDaily>();
        wrapper2.le("dat", DateUtils.getDate());
        agentExpandDailyService.remove(wrapper2);
        agentExpandDailyService.insertByDetail();
    }

    @Data
    class InnerUserAgent {
        Integer userId;
        Integer agentId;
        Integer agentLevel;
    }
}
